/*
 * this file is part of project
 * PIC4BBB
 *
 * Programmer for dsPIC33EP / PIC24EP
 * with BeagleBone BLACK
 * M. 2017
 *
 * Two variants are possible:
 * to use I2C pins (default is I2C2 is used for capes and also can communicate with others IO)
 * I2C1 will be used to communicate with PICs - fast and can be used several PICs on line (with different MCLR control pins)
 * PICs has I2C on PGE2 pins (PGED2-ASDA2 and PGEC2-ASCL2); alternative I2C2 of the PIC configs. in FPOR)
 *
 * or UART pins - simple, but slow and for one PIC only
 * PICs has possibilities for UART on PGE1 pins (PGED1-RP35, config. as Tx and PGEC1-RPI34, config. as Rx)
 *
 * This not uses pins for LCD cape of internal eMMC.
 * This is not cape, it can use capes I2C2, but after booting.
 * The software is library (Python).
 *
Free, open-source, fully featured demonstrative software.

THIS SOFTWARE IS PROVIDED,
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
PROVIDER BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

// System
#define reg(x) (*((volatile unsigned int *)(x)))
// Memory cell 32bit x = 0-2999 (12KB)
#define mem(x) reg(0x10000+(x<<2))

// addresses for PRU
#define SYSCFG  0x26004

#define GPIO0 0x44e07000
#define GPIO1 0x4804c000
#define GPIO2 0x481ac000
#define GPIO3 0x481ae000

#define GPIO_CLEARDATAOUT 0x190
#define GPIO_SETDATAOUT 0x194
#define GPIO_DATAOUT 0x138

volatile register unsigned int __R31, __R30; // system registers

// Helper macros
#define len(x) (sizeof(x)/sizeof(unsigned int))

#define LW4(x) (x & 0xFFFF) << 4
#define HW4(x) (x & 0xFFFF0000) >> 12

#define exec_cmd for (i=0; i<len(cmds); ++i) send(cmds[i]);
#define exec_rst send(0); send(0); send(0);\
                 send(0x40200);\
                 send(0); send(0); send(0);

// Type
//#define viaUART
#define viaI2C

#ifdef viaUART
    // P9_11 -> UART4_RXD and gpio0[30] (gpio pu)
    // PGED1 can be PIC Tx (RP35)
    #define PGED (1<<30)
    // P9_13 -> UART4_TXD and gpio0[31] (out)
    // PGEC1 can be PIC Rx (RPI34)
    #define PGEC (1<<31)
#else
    // P9_17 -> I2C1_SCL and gpio0[5] (out)
    // PGEC2 can be PIC ASCL2
    #define PGEC (1<<5)
    // P9_18 -> I2C1_SDA and gpio0[4] (gpio pu)
    // PGED2 can be PIC ASDA2
    #define PGED (1<<4)
#endif

// because all are in GPIO0 ->
#define sC reg(GPIO0 | GPIO_SETDATAOUT) = PGEC
#define cC reg(GPIO0 | GPIO_CLEARDATAOUT) = PGEC
#define pC sC; d_p; cC; d_p

// PGED is input only - PRU can't change direction!
#define dat (reg(GPIO0 | GPIO_DATAOUT) & PGED)

// PGEDout - P9_25 - pr1_pru0_pru_r30_7 (pruout)
// Additional pin for PGED output - open drain with FET
// it is connected to PGED input, which has PullUP, but may be used external resistor (make experiment)
// and because of the FET -> PGEDout is inverted =>
#define cD __R30 |= 1 << 7
#define sD __R30 &= ~(1 << 7)

// to read from dat (PGED), the PGEDout must be '0' -> inverted -> PGED on PIC is '1', because PullUP
// (else only '0' will be read)
#define D2in sD

// Select for pin for PIC MCLR
// for example P9_26 UART1_RXD and gpio0[14] (out)
// (UART1_RXD can't be used, because UART1_TXD is used for LCD cape)
#define MCLR (1<<14)
#define sR reg(GPIO0 | GPIO_SETDATAOUT) = MCLR
#define cR reg(GPIO0 | GPIO_CLEARDATAOUT) = MCLR

// delays (cycles are of 5ns)
#define d_us(x) __delay_cycles((x)*200)

// pulse delay - min 100ns =>
// 100/5 = 20 cycles
// (really min is 80ns, but period must be > 200ns,
// this will be ensured from the commands execution time -> extra delay)
// if in circuit returns errors - test with lower frequency (increase delay value) because of bad waveform
#define d_p __delay_cycles(30)

#define d_5ns(x) __delay_cycles(x)


// Send 24 bit command to PIC
// PIC reads data on FALLING edge of PGEC
void send(unsigned int x){
    unsigned int i;

    d_5ns(8); // P4a
    cD;
    for (i=0; i<4; ++i) {pC;} // send SIX
    d_5ns(8); // P4
    for (i=0; i<24; ++i){
        if (x & (1 << i)) sD; else cD;
        pC;
    }
}


// Receive 16bit VISI from PIC
// PIC outs data on RISING edge of PGEC
unsigned int rec(void){
    unsigned int i, out=0;

    d_5ns(8); // P4a
    sD; pC; cD; pC; pC; pC; // send REGOUT
    d_5ns(8); // P4
    for (i=0; i<8; ++i) {pC;} // CPU idle
    D2in;
    d_5ns(4); // P5
    for (i=0; i<16; ++i){
        sC;
        if (dat) out |= 1 << i;
        d_p; cC; d_p;
    }

    return out;
}


// Erasing whole memory
void erase(void){
    unsigned int i;

    static unsigned int cmds[] = {0x2400FA, 0x88394A, 0, 0,
                           0x200551, 0x883971, 0x200AA1, 0x883971, 0xA8E729, 0, 0, 0};
    exec_rst;
    exec_cmd;
    d_us(120000); // P11
}


// Write memory from addr (24 bit) with two commands data (24 bit)
// data0 -> MSB1:MSB0 :LSW0; data1 -> LSW1
void write(unsigned int addr, unsigned int data0, unsigned int data1){
    unsigned int i;

    static unsigned int cmds[] = {0x200FAC, 0x8802AC,
                                  0x200000, 0x200001, 0x200002, // here between '2' and '0' is: LSW0, MSB1:MSB0, LSW1
                                  0xEB0300, 0, 0xEB0380, 0, 0xBB0BB6, 0, 0, 0xBBDBB6, 0, 0, 0xBBEBB6, 0, 0, 0xBB1BB6, 0, 0,
                                  0x200003, 0x200004, 0x883953, 0x883964, // here between '2' and '3/4' is: 16 low addr, 8 high addr
                                  0x24001A, 0, 0x88394A, 0, 0,
                                  0x200551, 0x883971, 0x200AA1, 0x883971, 0xA8E729, 0, 0, 0, 0, 0};

    // prepare addr and data
    cmds[2] = 0x200000; cmds[3] = 0x200001; cmds[4] = 0x200002;
    cmds[21] = 0x200003; cmds[22] = 0x200004;

    cmds[2] |= LW4(data0); // LSW0
    cmds[3] |= HW4(data0); // MSB1:MSB0
    cmds[4] |= LW4(data1); // LSW1

    cmds[21] |= LW4(addr); // LW addr
    cmds[22] |= HW4(addr); // HW(HB) addr

    exec_rst;
    exec_cmd;
    d_us(2000); // P13
}


// Write 2x 16bit data (config. or UsetID) in memory addr (24 bit)
// Config. registers are 8bit, high 8B of 16 bit data must be 0xFF -> this function writes TWO config. regs. (read16 fill read the HB as 00!)
// do not modify Config data - the HEX file prepare properly - HB = 0xFF
// UserID is 16 bit data; TWO UserID cells are written with this function
// ATTENTION!: This write only in HIGHER memory - TABLEPAG is set (see first command), so use only for Config and user ID!!!
void write16(unsigned int addr, unsigned int data){
    unsigned int i;

    static unsigned int cmds[] = {0x200FAC, 0x8802AC,
                                  0x200000, 0x200001, // here between '2' and '0/1' is: LW data, and second HW data
                                  0xEB0180, 0, 0xBB1980, 0, 0, 0xBB0981, 0, 0,
                                  0x200004, 0x200005, 0x883954, 0x883965, // here between '2' and '4/5' is: 16 low addr, 8 high addr
                                  0x24001A, 0, 0x88394A, 0, 0,
                                  0x200551, 0x883971, 0x200AA1, 0x883971, 0xA8E729, 0, 0, 0, 0, 0};

    // prepare addr and data
    cmds[2] = 0x200000; cmds[3] = 0x200001;
    cmds[12] = 0x200004; cmds[13] = 0x200005;

    cmds[2] |= LW4(data); // first W
    cmds[3] |= HW4(data); // second W

    cmds[12] |= LW4(addr); // 16 bit low addr
    cmds[13] |= HW4(addr); // 8 bit high addr

    exec_rst;
    exec_cmd;
    d_us(2000); // P13
}


// Read memory from addr (24 bit) 4 commands data (24 bit) - returned as array
// reads in 16 bit packets in this order LSW0, MSB1:MSB0, LSW1
unsigned int *read(unsigned int addr){
    unsigned int i;
    unsigned int data[6]; // each is 16 bit effective -> 4 result-s (24bit codes) -> (data[5]) LSW3, MSB3:MSB2, LSW2, LSW1, MSB1:MSB0, LSW0 (data[0])
    static unsigned int result[4]; // [0]: MSB1:MSB0 :LSW0
                                   // [1]: LSW1
                                   // [2]: MSB3:MSB2 :LSW2
                                   // [3]: LSW3

    static unsigned int cmds[] = {0x200000, 0x8802A0, 0x200006,// here between '2' and '0/6' is: addr higher 8 bit and lower 16 bit

                                  0xEB0380, 0, 0xBA1B96, 0, 0, 0, 0, 0, 0xBADBB6, 0, 0, 0, 0, 0, 0xBADBD6, 0, 0, 0, 0, 0, 0xBA1BB6, 0, 0, 0, 0, 0,
                                  0xBA1B96, 0, 0, 0, 0, 0, 0xBADBB6, 0, 0, 0, 0, 0, 0xBADBD6, 0, 0, 0, 0, 0, 0xBA0BB6, 0, 0, 0, 0, 0};

    // prepare addr and data
    cmds[0] = 0x200000; cmds[2] = 0x200006;

    cmds[0] |= HW4(addr); // 8 bit high addr
    cmds[2] |= LW4(addr); // 16 bit lower addr

    exec_rst;
    exec_cmd;

    // Read data
    for (i=0; i<6; ++i){
        send(0x887C40+i); send(0);
        data[i] = rec();
        send(0);
    }

    result[0] = (data[1] << 16) | data[0];
    result[1] = data[2];

    result[2] = (data[4] << 16) | data[3];
    result[3] = data[5];

    return result;
}


// Read memory from addr (24 bit) 16 bit data
// if isCfg=1 return only low 8bit, but returns and HB - what is written, this reads -> FF is written, reads HB=0xFF
unsigned int read16(unsigned int addr, unsigned int isCfg){
    unsigned int i;

    static unsigned int cmds[] = {0x200000, 0x8802A0, 0x200006,// here between '2' and '0/6' is: addr higher 8 bit and lower 16 bit
                                  0xEB0380, 0, 0xBA5B96, 0, 0, 0, 0, 0,
                                  0x887C40, 0};

    cmds[5] = (isCfg) ? 0xBA5B96 : 0xBA1BB6; // if Config. reads - 8 bit, else UserID - 16 bit

    // prepare addr and data
    cmds[0] = 0x200000; cmds[2] = 0x200006;

    cmds[0] |= HW4(addr); // 8 bit high addr
    cmds[2] |= LW4(addr); // 16 bit lower addr

    exec_rst;
    exec_cmd;

    return rec();
}



int main(void) {
    unsigned int i, addr, cellNum;
    unsigned int *result;
    unsigned char isProgramming=1;

    union{
        unsigned char full;
        struct{
            unsigned char isWritten:1;
            unsigned char isError:1;
            enum{
                pdata = 0,
                id,
                config
            } type:2;
            unsigned char b4:1;
            unsigned char b5:1;
            unsigned char b6:1;
            unsigned char b7:1;
        };
    } ctrlB;


    reg(SYSCFG) &= ~(1<<4); // enable external memory access
    d_us(1000);

    // Prepare pin states
    cR; d_us(1); cC; cD;
    d_us(10000);

    // Enter ICSP
    // ==========
    sR;
    d_us(300); // P21
    cR;
    d_us(2000); // P18

    for (i=0; i<32; ++i){
        if (0x4D434851 & (0x80000000 >> i)) sD; else cD;
        pC;
    }
    cD;
    d_us(1); // P19
    sR;
    d_us(60000); // > P7 + 5*P1

    // 5 extra pulses
    for (i=0; i<5; ++i){pC;}
    d_us(10); // P4
    // Ready for commands

    // Programming
    // ===========
    erase();


    // Shared RAM - 32 bit cells 0 up to 2999
    // cell 0:
    // low 24 bits are the start addr. of the "page"
    // high 8 bits are flags:
    // b0 - '0' - to write - PRU writes and checks the written and set this bit to '1' = "end"
    // b1 - '0' - no errors, if PRU detect error in check of the programmed - set this bit to '1'
    // b3:b2 - '00' - program data; '01' - user ID; '10' - config. data
    //
    // from cell 1:
    // if b3:b2 = '00' =>
    // data cells are PIC Program Memory in format:
    // [0]: MSB1:MSB0 :LSW0
    // [1]: LSW1
    // [2]: MSB3:MSB2 :LSW2
    // [3]: LSW3
    // and so on...
    // if the HW of cell with 16 bit data only (LSW 1, 3...)
    // is 0xFFFF (main application set this) - this is the last from the page
    //
    // so in one page:
    // 1 - "system" cell
    // 2999 - data cells
    // two data cells consists of two commands (2x 24bit), which are written at once =>
    // 1499 packets by two commands can be written in one "page"
    // but for checking, which returns two such packets, i.e. 4 commands of 24 bit
    // we use cell number, divisible by 4 =>
    // 749 packets by 4 codes are used in one "page" -> 2996 cells of 32 bit
    // last 3 cells remain "reserved"
    //
    // in short:
    // cell 0: system
    // cell 1-2996 - data
    // cell 2997, 2998 and 2999 - reserved
    // the biggest PIC controller has up to approx. 175110 codes (24 bit)
    // or 43777,5 packets by 4 codes (24 bit)
    // 59 "pages" (not full) can be used!
    //
    // if b3:b2 = '01' - User ID - as for the config., but isCfg is '0'
    // if b3:b2 = '10' - config., set isCfg in read16 to '1'
    // data consists of two cells by 16bit (this are written at once).
    // These are 1 page, have no end marker.
    // User ID are always 4 cells
    // Config. are always 6 cells.
    // Seems that even if locked, the Configs can be verified?!
    // Exits the ICSP - end the programming

    while (isProgramming){
        while(mem(0) & 0x1000000); // wait new write command (b0)

        addr = mem(0) & 0xFFFFFF; // read start address
        ctrlB.full = mem(0) >> 24; // read control Byte (bit 24-31)

        cellNum = 1;
        // Writing
        // -------
        switch (ctrlB.type){
            // b3:b2
            case pdata:
                // Program Memory
                while(1){
                    write(addr, mem(cellNum), mem(cellNum+1));
                    if ((mem(cellNum+1) >> 16) == 0xFFFF) break;
                    addr += 4; cellNum += 2;
                }
                break;
            case id:
                // user ID
                write16(0x800FF8, (mem(2)<<16) | mem(1)); // this writes 0x800FF8 (mem(1)) and 0x800FFA (mem(2))
                write16(0x800FFC, (mem(4)<<16) | mem(3)); // this writes 0x800FFC (mem(3)) and 0x800FFE (mem(4))
                break;
            case config:
                write16(addr, (mem(2)<<16) | mem(1));   // FICD and FPOR
                write16(addr+4, (mem(4)<<16) | mem(3)); // FWDT and FOSC
                write16(addr+8, (mem(6)<<16) | mem(5)); // FOSCSEL and FGS
            default: break;
        }

        // Reading (check)
        // -------
        addr = mem(0) & 0xFFFFFF; // read start address

        cellNum = 1;
        switch (ctrlB.type){
            // b3:b2
            case pdata:
                // Program Memory
                while(1){
                    result = read(addr);

                    if ((result[0] != mem(cellNum)) ||
                        (result[1] != mem(cellNum+1)) ||
                        (result[2] != mem(cellNum+2)) ||
                        (result[3] != (mem(cellNum+3) & 0xFFFF))){ // here remove eventual "stop marker" of HW
                        ctrlB.isError = 1; ctrlB.isWritten = 1;// set Error and End
                        break;
                    }
                    if ((mem(cellNum+3) >> 16) == 0xFFFF){
                        ctrlB.isWritten = 1; // set End (successful)
                        break;
                    }
                    addr += 8; cellNum += 4;
                }
                break;
            case id:
                if ((read16(0x800FF8, 0) != (mem(1) & 0xFFFF)) ||
                    (read16(0x800FFA, 0) != (mem(2) & 0xFFFF)) ||
                    (read16(0x800FFC, 0) != (mem(3) & 0xFFFF)) ||
                    (read16(0x800FFE, 0) != (mem(4) & 0xFFFF))) ctrlB.isError = 1;
                ctrlB.isWritten = 1;
                break;
            case config:
                if ((read16(addr, 1) != (mem(1) & 0xFF)) ||
                    (read16(addr+2, 1) != (mem(2) & 0xFF)) ||
                    (read16(addr+4, 1) != (mem(3) & 0xFF)) ||
                    (read16(addr+6, 1) != (mem(4) & 0xFF)) ||
                    (read16(addr+8, 1) != (mem(5) & 0xFF)) ||
                    (read16(addr+10, 1) != (mem(6) & 0xFF))){ctrlB.isError = 1; ctrlB.isWritten = 1; break;}

                ctrlB.isWritten = 1; // set End
                isProgramming = 0; // Out of the ICSP
                break;
            default: break;
        }
        mem(0) = ((unsigned int)ctrlB.full << 24) | (mem(0) & 0xFFFFFF); // return reply to main application
    }

    // Exit ICSP
    // =========
    exec_rst;
    // Prepare pin states - to can start work the PIC
    cR; d_us(1); cC; cD;
    d_us(100000); // ensure PIC reset
    sR; // This will run the PIC program

    // Send interrupt to host and halt
    __R31 = 35;
    __halt();

    return 0;
}
